@@ -132,12 +132,15 @@ struct options {
     int     numspeed;			/* number of baud rates to try */
     int     speeds[MAX_SPEED];		/* baud rates to be tried */
     char   *tty;			/* name of tty */
+    char   *initstring;			/* modem init string */
 };
 
 #define	F_PARSE		(1<<0)		/* process modem status messages */
 #define	F_ISSUE		(1<<1)		/* display /etc/issue */
 #define	F_RTSCTS	(1<<2)		/* enable RTS/CTS flow control */
 #define F_LOCAL		(1<<3)		/* force local */
+#define F_INITSTRING    (1<<4)		/* initstring is set */
+#define F_WAITCRLF	(1<<5)		/* wait for CR or LF */
 
 /* Storage for things detected while the login name was read. */
 
@@ -159,6 +162,49 @@ struct chardata init_chardata = {
     0,					/* always filled in at runtime */
 };
 
+struct Speedtab {
+    long    speed;
+    int     code;
+};
+
+static struct Speedtab speedtab[] = {
+    50, B50,
+    75, B75,
+    110, B110,
+    134, B134,
+    150, B150,
+    200, B200,
+    300, B300,
+    600, B600,
+    1200, B1200,
+    1800, B1800,
+    2400, B2400,
+    4800, B4800,
+    9600, B9600,
+#ifdef	B19200
+    19200, B19200,
+#endif
+#ifdef	B38400
+    38400, B38400,
+#endif
+#ifdef	EXTA
+    19200, EXTA,
+#endif
+#ifdef	EXTB
+    38400, EXTB,
+#endif
+#ifdef B57600
+    57600, B57600,
+#endif
+#ifdef B115200
+    115200, B115200,
+#endif
+#ifdef B230400
+    230400, B230400,
+#endif
+    0, 0,
+};
+
 #define P_(s) ()
 void parse_args P_((int argc, char **argv, struct options *op));
 void parse_speeds P_((struct options *op, char *arg));
@@ -212,7 +258,7 @@ main(argc, argv)
 #endif
 
 #ifdef DEBUGGING
-	dbf = fopen("/dev/tty1", "w");
+	dbf = fopen("/dev/ttyp0", "w");
 
 	{	int i;
 	
@@ -236,6 +282,7 @@ main(argc, argv)
     update_utmp(options.tty);
 #endif
 
+    debug("calling open_tty\n");
     /* Open the tty as standard { input, output, error }. */
     open_tty(options.tty, &termio, options.flags & F_LOCAL);
 
@@ -248,21 +295,45 @@ main(argc, argv)
 	}
 #endif
     /* Initialize the termio settings (raw mode, eight-bit, blocking i/o). */
-
+    debug("calling termio_init\n");
     termio_init(&termio, options.speeds[FIRST_SPEED], options.flags & F_LOCAL);
-	
-    /* Optionally detect the baud rate from the modem status message. */
 
+    /* write the modem init string and DON'T flush the buffers */
+    if (options.flags & F_INITSTRING) {
+	debug("writing init string\n");
+	write(1, options.initstring, strlen(options.initstring));
+    }
+
+    if (!(options.flags & F_LOCAL)) {
+	/* go to blocking write mode unless -L is specified */
+	fcntl(1, F_SETFL, fcntl(1, F_GETFL, 0) & ~O_NONBLOCK);
+    }
+
+    /* Optionally detect the baud rate from the modem status message. */
+    debug("before autobaud\n");
     if (options.flags & F_PARSE)
 	auto_baud(&termio);
 
     /* Set the optional timer. */
-
     if (options.timeout)
 	(void) alarm((unsigned) options.timeout);
- 
-    /* Read the login name. */
 
+    /* optionally wait for CR or LF before writing /etc/issue */
+    if (options.flags & F_WAITCRLF) {
+	char ch;
+
+	debug("waiting for cr-lf\n");
+	while(read(0, &ch, 1) == 1) {
+	    ch &= 0x7f;   /* strip "parity bit" */
+#ifdef DEBUGGING
+	    fprintf(dbf, "read %c\n", ch);
+#endif
+	    if (ch == '\n' || ch == '\r') break;
+	}
+    }
+
+    /* Read the login name. */
+    debug("reading login name\n");
     while ((logname = get_logname(&options, &chardata, &termio)) == 0)
 	next_speed(&termio, &options);
 
@@ -298,8 +369,48 @@ parse_args(argc, argv, op)
     extern int optind;			/* getopt */
     int     c;
 
-    while (isascii(c = getopt(argc, argv, "Lhil:mt:"))) {
+    while (isascii(c = getopt(argc, argv, "I:Lhil:mt:w"))) {
 	switch (c) {
+	case 'I':
+	    if (!(op->initstring = malloc(strlen(optarg)))) {
+		error("can't malloc initstring");
+		break;
+	    }
+	    {
+		char ch, *p, *q;
+		int i;
+
+		/* copy optarg into op->initstring decoding \ddd
+		   octal codes into chars */
+		q = op->initstring;
+		p = optarg;
+		while (*p) {
+		    if (*p == '\\') {		/* know \\ means \ */
+			p++;
+			if (*p == '\\') {
+			    ch = '\\';
+			    p++;
+			} else {		/* handle \000 - \177 */
+			    ch = 0;
+			    for (i = 1; i <= 3; i++) {
+				if (*p >= '0' && *p <= '7') {
+				    ch <<= 3;
+				    ch += *p - '0';
+				    p++;
+				} else 
+				  break;
+			    }
+			}
+			*q++ = ch;
+		    } else {
+			*q++ = *p++;
+		    }
+		}
+		*q = '\0';
+	    }
+	    op->flags |= F_INITSTRING;
+	    break;
+
 	case 'L':				/* force local */
  	    op->flags |= F_LOCAL;
  	    break;
@@ -319,6 +430,9 @@ parse_args(argc, argv, op)
 	    if ((op->timeout = atoi(optarg)) <= 0)
 		error("bad timeout value: %s", optarg);
 	    break;
+	case 'w':
+	    op->flags |= F_WAITCRLF;
+	    break;
 	default:
 	    usage();
 	}
@@ -475,7 +589,8 @@ open_tty(tty, tp, local)
 	(void) close(0);
 	errno = 0;				/* ignore close(2) errors */
 
-	if (open(tty, (local ? O_RDWR|O_NONBLOCK : O_RDWR), 0) != 0)
+	debug("open(2)\n");
+	if (open(tty, O_RDWR|O_NONBLOCK, 0) != 0)
 	    error("/dev/%s: cannot open as standard input: %m", tty);
 
     } else {
@@ -490,7 +605,7 @@ open_tty(tty, tp, local)
     }
 
     /* Set up standard output and standard error file descriptors. */
-
+    debug("duping\n");
     if (dup(0) != 1 || dup(0) != 2)		/* set up stdout and stderr */
 	error("%s: dup problem: %m", tty);	/* we have a problem */
 
@@ -537,8 +652,8 @@ termio_init(tp, speed, local)
      * lateron.
      */
 #ifdef linux
-	/* flush input and output queues, important for modems! */
-	(void) ioctl(0, TCFLSH, 2);
+    /* flush input and output queues, important for modems! */
+    (void) ioctl(0, TCFLSH, TCIOFLUSH);
 #endif
 
     tp->c_cflag = CS8 | HUPCL | CREAD | speed;
@@ -550,9 +665,10 @@ termio_init(tp, speed, local)
     tp->c_cc[VMIN] = 1;
     tp->c_cc[VTIME] = 0;
     (void) ioctl(0, TCSETA, tp);
-    if (local) {
-	(void) fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) & ~O_NDELAY);
-    }
+
+    /* go to blocking input even in local mode */
+    fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) & ~O_NONBLOCK);
+
     debug("term_io 2\n");
 }
 
@@ -707,13 +823,15 @@ do_prompt(op, tp)
 
 		  case 'b':
 		    {
-		      char *zpeedz[] = { "0", "50", "75", "110", "134.5",
-					 "150", "200", "300", "600", "1200",
-					 "1800", "2400", "4800", "9600", 
-					 "19200", "38400" };
+			int i;
 
-		      (void) printf ("%s", zpeedz[tp->c_cflag & CBAUD]);
-		      break;
+			for (i = 0; speedtab[i].speed; i++) {
+			    if (speedtab[i].code == (tp->c_cflag & CBAUD)) {
+				printf("%ld", speedtab[i].speed);
+				break;
+			    }
+			}
+			break;
 		    }
 		  case 'u':
 		  case 'U':
@@ -796,7 +914,7 @@ char   *get_logname(op, cp, tp)
     /* Flush pending input (esp. after parsing or switching the baud rate). */
 
     (void) sleep(1);
-    (void) ioctl(0, TCFLSH, (struct termio *) 0);
+    (void) ioctl(0, TCFLSH, TCIFLUSH);
 
     /* Prompt for and read a login name. */
 
@@ -971,38 +1089,6 @@ int
 bcode(s)
      char   *s;
 {
-    struct Speedtab {
-	long    speed;
-	int     code;
-    };
-    static struct Speedtab speedtab[] = {
-	50, B50,
-	75, B75,
-	110, B110,
-	134, B134,
-	150, B150,
-	200, B200,
-	300, B300,
-	600, B600,
-	1200, B1200,
-	1800, B1800,
-	2400, B2400,
-	4800, B4800,
-	9600, B9600,
-#ifdef	B19200
-	19200, B19200,
-#endif
-#ifdef	B38400
-	38400, B38400,
-#endif
-#ifdef	EXTA
-	19200, EXTA,
-#endif
-#ifdef	EXTB
-	38400, EXTB,
-#endif
-	0, 0,
-    };
     struct Speedtab *sp;
     long    speed = atol(s);
 
@@ -1017,15 +1103,11 @@ bcode(s)
 void
 usage()
 {
-#if defined(SYSV_STYLE) && !defined(linux)
     static char msg[] =
-    "[-i] [-l login_program] [-m] [-t timeout] line baud_rate,...";
-#else
-    static char msg[] =
-    "[-h] [-l login_program] [-m] [-t timeout] baud_rate,... line";
-#endif
+    "[-hiLmw] [-l login_program] [-t timeout] [-I initstring] baud_rate,... line [termtype]\nor\t[-hiLmw] [-l login_program] [-t timeout] [-I initstring] line baud_rate,... [termtype]";
 
-    error("usage: %s %s", progname, msg);
+    fprintf(stderr, "Usage: %s %s\n", progname, msg);
+    exit(1);
 }
 
 /* error - report errors to console or syslog; only understands %s and %m */
